import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:flutter_secured_storage/flutter_secured_storage.dart';
import 'package:m4e_sdk/m4e_sdk.dart';
import 'package:sdk_example/services/auth_service.dart';

enum AuthRepositoryState { initial, loading, success, error }

class AuthRepository with ChangeNotifier {
  final _authService = AuthService();
  final _storage = FlutterSecuredStorage();

  static AuthRepositoryState _state;
  AuthRepositoryState get state => _state ?? AuthRepositoryState.initial;

  static M4eAuthPhoneNumberForm _phoneNumberForm;
  M4eAuthPhoneNumberForm get phoneNumberForm => _phoneNumberForm;

  static M4eException _error;
  M4eException get error => _error;

  static M4eAuthToken _m4eAuthToken;
  M4eAuthToken get m4eAuthToken => _m4eAuthToken;

  static M4eUser _m4eUser;
  M4eUser get m4eUser => _m4eUser;

  Future<void> isAuthenticated() async {
    try {
      _state = AuthRepositoryState.loading;
      notifyListeners();

      // cache m4e-user
      final _m4eUserString = await _storage.read(key: 'M4E_TOKEN');

      if (_m4eUserString != null) {
        _m4eAuthToken = M4eAuthToken.fromJson(
            json.decode(_m4eUserString) as Map<String, dynamic>);
      }

      _state = AuthRepositoryState.success;
      notifyListeners();
    } on M4eException catch (e) {
      _error = e;
      _state = AuthRepositoryState.error;
      notifyListeners();
    }
  }

  Future<void> signInWithPhoneNumber(
      {@required M4eAuthPhoneNumberForm phoneNumberForm}) async {
    try {
      _state = AuthRepositoryState.loading;
      _phoneNumberForm = phoneNumberForm;
      notifyListeners();

      // sign-in with phone number
      await _authService.signInWithPhoneNumber(phoneNumberForm);

      _state = AuthRepositoryState.success;
      notifyListeners();
    } on M4eException catch (e) {
      _error = e;
      _state = AuthRepositoryState.error;
      notifyListeners();
    }
  }

  Future<void> verifyOTPCode({@required String otpCode}) async {
    try {
      _state = AuthRepositoryState.loading;
      notifyListeners();

      final _m4eAuthOTPVerificationForm =
          M4eAuthOTPVerificationForm(otpCode: otpCode);

      // verify otp-code
      _m4eAuthToken =
          await _authService.verifyOTPCode(_m4eAuthOTPVerificationForm);

      // cache m4e-token
      _storage.write(
          key: 'M4E_TOKEN', value: json.encode(_m4eAuthToken.toJson()));

      _state = AuthRepositoryState.success;
      notifyListeners();
    } on M4eException catch (e) {
      _error = e;
      _state = AuthRepositoryState.error;
      notifyListeners();
    }
  }

  Future<void> getAuthentiicatedUserAccount() async {
    try {
      _state = AuthRepositoryState.loading;
      notifyListeners();

      // fetch user-account from cache
      final _userAccountString = await _storage.read(key: 'M4E_USER');

      if (_userAccountString != null) {
        return M4eUser.fromJson(json.decode(_userAccountString));
      }

      // fetch account from server
      _m4eUser = await _authService.getAuthenticatedUserAccount();

      // cache m4e-user
      _storage.write(key: 'M4E_USER', value: json.encode(_m4eUser.toJson()));

      _state = AuthRepositoryState.success;
      notifyListeners();
    } on M4eException catch (e) {
      _error = e;
      _state = AuthRepositoryState.error;
      notifyListeners();
    }
  }

  Future<void> createUserAccount({@required M4eUserForm m4eUserForm}) async {
    try {
      _state = AuthRepositoryState.loading;
      notifyListeners();

      // create account
      await _authService.createUserAccount(m4eUserForm);

      _state = AuthRepositoryState.success;
      notifyListeners();
    } on M4eException catch (e) {
      _error = e;
      _state = AuthRepositoryState.error;
      notifyListeners();
    }
  }
}
