import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../models/models.dart';
import '../../data/services/utils/utils.dart';

/// Accessing the instantiated `TransactionApi` object
/// is done by passing its type to the GetIt instance
///
/// Example use of `TransactionApi`:
///
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<TransactionApi>().initiateTransaction(InternalTransactionForm transactionForm);
/// ```
///
///
/// To create an transaction, you can do
/// ```
/// final transactionForm = InternalTransactionForm(from: fromWallet, to: toWallet, amount: amount ...)
/// ```
///
///
abstract class TransactionApi {
  /// Initiates a payment transaction by sending [InternalTransactionForm] to m4e.
  ///
  /// Transaction status is `pending` until m4e returns a response which can be `M4eTransactionResult`
  /// or `ServerException` on success and failure cases respectively.
  ///
  /// Throws `ServerException` with code `FAILED_TRANSACTION` if m4e responds with an error which may
  /// be an unauthorized request (user does not have permission to pay into
  /// the receiving wallet), wallet not found, etc.
  ///
  Future<M4eTransactionResult> initiateTransaction(
      InternalTransactionForm internalTransactionForm);

  /// Returns m4e authenticated user's `TransactionList` from m4e server
  ///
  /// Throws `ServerException` with code `FAILED_TRANSACTION` if m4e responds with an error which may
  /// be an unauthorized request (user does not have permission to pay into
  /// the receiving wallet), wallet not found, etc.
  ///
  Future<M4eTransactionList> getUserTransactions(M4eUniqueId userId);

  /// Returns m4e authenticated user's wallet `TransactionList` from m4e server
  ///
  /// Throws `ServerException` with code `FAILED_TRANSACTION` if m4e responds with an error which may
  /// be an unauthorized request (user does not have permission to pay into
  /// the receiving wallet), wallet not found, etc.
  ///
  Future<M4eTransactionList> getWalletTransactions(M4eUniqueId walletId);
}

class TransactionApiImpl implements TransactionApi {
  final Dio _client;

  TransactionApiImpl({
    @required Dio client,
    @required M4eLoggingInterceptors loggingInterceptor,
    @required M4eHeaderInterceptors headerInterceptor,
  })  : assert(client != null),
        assert(loggingInterceptor != null),
        assert(headerInterceptor != null),
        _client = client {
    _client?.interceptors?.add(loggingInterceptor);
    _client?.interceptors?.add(headerInterceptor);
  }

  Future<M4eTransactionResult> initiateTransaction(
      InternalTransactionForm transactionForm) async {
    try {
      final _response = await _client.put(
        '/transactions/${M4eUniqueId().toString()}',
        data: json.encode(transactionForm.toJson()),
      );

      return M4eTransactionResult.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }

  @override
  Future<M4eTransactionList> getUserTransactions(M4eUniqueId userId) async {
    try {
      final _fromDate = DateTime(SettingValues.kMinimumYearSpecification)
          .toTimestampSinceEpoch();
      final _epoch = DateTime.now().toTimestampSinceEpoch();

      final _response = await _client.get(
          '/transactions?from=$_fromDate&to=$_epoch&at=$_epoch&entityId=${userId.toString()}&access=${EntityWalletAccess.pay}');

      return M4eTransactionList.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }

  @override
  Future<M4eTransactionList> getWalletTransactions(M4eUniqueId walletId) async {
    try {
      final _fromDate = DateTime(SettingValues.kMinimumYearSpecification)
          .toTimestampSinceEpoch();
      final _epoch = DateTime.now().toTimestampSinceEpoch();

      final _response = await _client.get(
          '/transactions?from=$_fromDate&to=$_epoch&at=$_epoch&walletIds=${walletId.toString()}');

      return M4eTransactionList.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }
}
