import 'package:data_connection_checker/data_connection_checker.dart';

import '../../m4e_sdk_initialize.dart';
import '../api/apis.dart';
import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../models/models.dart';

/// Exposes SDK Auth Api methods
///
/// Accessing the instantiated `M4eAuthService` methods
/// is done by:
///
/// ```
/// final _authService = M4eAuthService();
///
/// final _isAuthenticated = await _authService.isAuthenticated();
/// ```
///
class M4eAuthService {
  final AuthApi _authApi;
  final DataConnectionChecker _connectionChecker;

  M4eAuthService({
    AuthApi authApi,
    DataConnectionChecker connectionChecker,
  })  : _authApi = authApi ?? m4eSDK<AuthApi>(),
        _connectionChecker =
            connectionChecker ?? m4eSDK<DataConnectionChecker>();

  /// Returns `AuthToken` if user is authenticated
  ///
  /// Throws `AuthException` with `NO_AUTH_USER` error
  /// code if no user is authenticated
  ///
  Future<M4eAuthToken> isAuthenticated() async {
    try {
      final _token = await _authApi.getAuthTokenFromCache();

      return _token;
    } on M4eAuthException catch (_) {
      rethrow;
    }
  }

  /// Returns `M4eAuthPhoneNumberForm` credentials
  /// used for m4e authentication
  ///
  /// Throws `AuthException` with `NO_CACHED_AUTH_CREDENTIAL` error
  /// code if no authentication credential exists
  ///
  Future<M4eAuthPhoneNumberForm> getAuthCredentials() async {
    try {
      final _credentials = await _authApi.getUserCredentialsFromCache();

      return _credentials;
    } on M4eAuthException catch (_) {
      rethrow;
    }
  }

  /// Requests `access-token` from m4e using existing `refresh-token`
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eAuthToken> refreshAccessToken() async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _existingToken = await _authApi.getAuthTokenFromCache();

        final _token = await _authApi.refreshAccessToken(_existingToken);

        // cache new token
        await _authApi.cacheAuthToken(_token);

        return _token;
      } catch (e) {
        rethrow;
      }
    }

    throw M4eExceptionMessages.kNoInternetConnectionException;
  }

  /// Takes in `M4eAuthPhoneNumberForm` parameter to send
  /// verification-code (OTP) from m4e to provided phone number
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  /// code if device is not connected to the internet
  ///
  Future<M4eUniqueId> verifyPhoneNumber(
      M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _userId =
            await _authApi.verifyPhoneNumber(m4eAuthPhoneNumberForm);

        final _credentials = m4eAuthPhoneNumberForm.copyWith(id: _userId);
        await _authApi.cacheUserCredentials(_credentials);

        return _userId;
      } catch (e) {
        rethrow;
      }
    }

    throw M4eExceptionMessages.kNoInternetConnectionException;
  }

  /// Takes in `M4eAuthOTPVerificationForm` parameter
  /// and returns `AuthToken` on success call
  ///
  /// Throws `AuthException` with `NO_CACHED_AUTH_CREDENTIAL` error
  /// code if no authentication credential exists
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  /// code if device is not connected to the internet
  ///
  Future<M4eAuthToken> verifyOTPCode(
      M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _m4eAuthPhoneNumberCredential =
            await _authApi.getUserCredentialsFromCache();

        final _authToken = await _authApi.verifyOTPCode(
            _m4eAuthPhoneNumberCredential, m4eAuthOTPVerificationForm);

        await _authApi.cacheAuthToken(_authToken);
        return _authToken;
      } catch (e) {
        rethrow;
      }
    }

    throw M4eExceptionMessages.kNoInternetConnectionException;
  }

  /// Signout authenticated m4e user
  Future<M4eUnitImpl> signout() async {
    await _authApi.signout();
    return unitImpl;
  }
}
