import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';
import 'package:shortid/shortid.dart';

import '../../../core/core.dart';
import '../../../m4e_forms/m4e_forms.dart';
import '../../../models/models.dart';
import '../../api/apis.dart';

class TransactionServiceUtil {
  final WalletApi _walletApi;

  TransactionServiceUtil({
    @required WalletApi walletApi,
  })  : assert(walletApi != null),
        _walletApi = walletApi;

  /// map [M4eTransactionForm] to [InternalTransactionForm] for m4e-sdk internal transactions
  Future<InternalTransactionForm> m4eProviderTransactionMapper(
      M4eTransactionForm transaction) async {
    final _reciepient = transaction.to as M4eWalletProvider;

    return InternalTransactionForm(
      from: transaction.from,
      to: _reciepient.wallet,
      amount: transaction.amount,
      at: transaction.at,
      meta: transaction.meta,
    );
  }

  /// map [M4eTransactionForm] to [InternalTransactionForm] for m4e-sdk internal transactions
  ///
  /// query and run query for reciepient wallet-label on their phone number
  ///
  /// creates a new wallet-label with receipient's phone number if no
  /// wallet-label associated with recipient phone number is returned and
  /// initiate transaction with the newly created wallet
  ///
  /// in cases where a wallet-label already exists for receipient's phone
  /// number, proceed with initiating a transaction with returned wallet
  Future<InternalTransactionForm> mobileMoneyProviderTransactionMapper(
      M4eUniqueId userId, M4eTransactionForm transaction) async {
    try {
      final _reciepient = transaction.to as MobileMoneyWalletProvider;

      final _composedWalletSearchForm = MobileMoneyProviderSearchForm(
        provider: _reciepient.provider,
        phoneNumber: _reciepient.phoneNumber,
      );

      final _runQueryUniqueId =
          await _walletApi.queryUserWallets(_composedWalletSearchForm);

      // returns list of wallets gotten from running search-returned id
      final _walletList = await _walletApi.runSearchQuery(_runQueryUniqueId);

      if (_walletList.wallets.isEmpty) {
        /// convert to wallet-form-type to create wallet
        final _toWalletFormType = _reciepient.toWalletForm(
          walletId: M4eUniqueId(),
          label: shortid.generate(),
          currencyType: transaction.amount.currency,
        );

        /// create wallet with transaction currency-type
        await _walletApi.createWallet(userId, _toWalletFormType);

        /// convert to wallet-type to make transaction
        final _toWalletType = M4eWallet(
          id: _toWalletFormType.id,
          walletInfo: M4eWalletInfo(
            label: _toWalletFormType.label,
            currency: _toWalletFormType.currency,
            type: _toWalletFormType.type,
          ),
        );

        return InternalTransactionForm(
          from: transaction.from,
          to: _toWalletType,
          amount: transaction.amount,
          at: transaction.at,
          meta: transaction.meta,
        );
      } else {
        /// existing wallet was returned
        /// map the first wallet-object returned to the transaction-to wallet
        ///
        final _wallet = _walletList.wallets.first;

        return InternalTransactionForm(
          from: transaction.from,
          to: _wallet,
          amount: transaction.amount,
          at: transaction.at,
          meta: transaction.meta,
        );
      }
    } catch (_) {
      rethrow;
    }
  }
}

/// Form parameters required for m4e-sdk transactions
///
class InternalTransactionForm extends Equatable {
  final M4eWallet from;
  final M4eWallet to;
  final M4eAmount amount;
  final DateTime at;
  final Map<String, dynamic> meta;

  const InternalTransactionForm._({
    @required this.from,
    @required this.to,
    @required this.amount,
    @required this.at,
    @required this.meta,
  });

  factory InternalTransactionForm({
    @required M4eWallet from,
    @required M4eWallet to,
    @required M4eAmount amount,
    @required DateTime at,
    @required Map<String, dynamic> meta,
  }) {
    assert(from != null, '_TransactionForm [from] wallet should not be null');
    assert(to != null, '_TransactionForm [to] wallet should not be null');
    assert(from.id != to.id,
        '_TransactionForm [from] and [to] wallets should not be the same');
    assert(amount != null, '_TransactionForm [amount] should not be null');
    assert(at != null, '_TransactionForm [at] should not be null');
    assert(meta != null, '_TransactionForm [meta] should not be null');

    return InternalTransactionForm._(
        from: from, to: to, amount: amount, at: at, meta: meta);
  }

  Map<String, dynamic> toJson() {
    final flow = {
      'from': this.from.toJson(),
      'to': this.to.toJson(),
      'amount': this.amount.toJson(),
      'at': this.at.toTimestampSinceEpoch()
    };

    // TODO: Transaction is a test object. Remove when not needed
    final _transactionType = this.from.walletInfo.type == M4eWalletType.M4E &&
            this.to.walletInfo.type == M4eWalletType.M4E
        ? M4eTransactionType.cashIn
        : M4eTransactionType.cashOut;

    return {
      'flow': flow,
      'txType': _transactionType.toString(),
      'meta': this.meta,
    };
  }

  @override
  List<Object> get props => [
        from,
        to,
        amount,
        at,
        meta,
      ];

  @override
  String toString() {
    return '_TransactionForm(from: $from, to: $to, amount: $amount, at: $at, meta: $meta)';
  }
}
