import 'package:data_connection_checker/data_connection_checker.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:m4e_sdk/src/core/core.dart';
import 'package:m4e_sdk/src/data/api/apis.dart';
import 'package:m4e_sdk/src/data/services/services.dart';
import 'package:m4e_sdk/src/m4e_forms/m4e_forms.dart';
import 'package:m4e_sdk/src/models/models.dart';
import 'package:matcher/matcher.dart';
import 'package:mockito/mockito.dart';

import 'test_models/wallet_test_model.dart';

class MockUserApi extends Mock implements UserApi {}

class MockWalletApi extends Mock implements WalletApi {}

class MockPermissionApi extends Mock implements PermissionApi {}

class MockDataConnectionChecker extends Mock implements DataConnectionChecker {}

main() {
  M4eWalletService _m4eWalletService;

  MockUserApi _mockUserApi;
  MockWalletApi _mockWalletApi;
  MockPermissionApi _mockPermissionApi;
  MockDataConnectionChecker _mockDataConnectionChecker;

  setUp(() {
    _mockUserApi = MockUserApi();
    _mockWalletApi = MockWalletApi();
    _mockPermissionApi = MockPermissionApi();
    _mockDataConnectionChecker = MockDataConnectionChecker();

    _m4eWalletService = M4eWalletService(
      userApi: _mockUserApi,
      walletApi: _mockWalletApi,
      permissionApi: _mockPermissionApi,
      connectionChecker: _mockDataConnectionChecker,
    );
  });

  group('m4e-wallet-test | ', () {
    group('getWallets | ', () {
      test('should verify device connectivity status', () async {
        // arrange
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
        when(_mockWalletApi.queryUserWallets(any))
            .thenAnswer((_) async => WalletTestModel.uniqueId);
        when(_mockWalletApi.runSearchQuery(any))
            .thenAnswer((_) async => WalletTestModel.tWalletList);

        // act
        await _m4eWalletService.getWallets(
            currentUserId: WalletTestModel.uniqueId);

        // assert
        verify(_mockDataConnectionChecker.hasConnection);
      });

      group('isConnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => true);
          when(_mockWalletApi.queryUserWallets(any))
              .thenAnswer((_) async => WalletTestModel.uniqueId);
          when(_mockWalletApi.runSearchQuery(any))
              .thenAnswer((_) async => WalletTestModel.tWalletList);
        });

        group('current-user-wallet', () {
          setUp(() {
            when(_mockUserApi.getM4eUser(any))
                .thenAnswer((_) async => WalletTestModel.m4eUser);
          });

          test(
              'should call [M4eUserApi] with [M4ePhoneNumber] as parameter to retrive [M4eUser]',
              () async {
            // act
            await _m4eWalletService.getWallets(
                currentUserId: WalletTestModel.uniqueId,
                otherUserPhoneNumber: WalletTestModel.phoneNumber);

            // assert
            verify(_mockUserApi.getM4eUser(WalletTestModel.phoneNumber));
          });

          test('should return [M4eWalletList] on success call', () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => WalletTestModel.uniqueId);

            // act
            final _result = await _m4eWalletService.getWallets(
                currentUserId: WalletTestModel.uniqueId);

            // assert
            expect(_result, equals(WalletTestModel.tWalletList));
          });

          test(
              'should return [WalletApi.queryUserWallets] with appropriate parameters (query with single-search-clause when [currentUserId] is passed in)',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => WalletTestModel.uniqueId);

            // act
            final _result = await _m4eWalletService.getWallets(
                currentUserId: WalletTestModel.uniqueId);

            // assert
            expect(_result, equals(WalletTestModel.tWalletList));
            verify(_mockWalletApi.queryUserWallets(
                WalletTestModel.singleSearchFormTypeCollection));
          });

          test(
              'should return [WalletApi.queryUserWallets] with appropriate parameters (query with single-search-clause when passed in [currentUserId] equals [otherUserId])',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => WalletTestModel.uniqueId);

            // act
            final _result = await _m4eWalletService.getWallets(
                currentUserId: WalletTestModel.uniqueId,
                otherUserPhoneNumber: WalletTestModel.phoneNumber);

            // assert
            expect(_result, equals(WalletTestModel.tWalletList));
            verify(_mockWalletApi.queryUserWallets(
                WalletTestModel.singleSearchFormTypeCollection));
          });

          test('should call [WalletApi.runSearchQuery] after creating query',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => WalletTestModel.uniqueId);

            // act

            await _m4eWalletService.getWallets(
                currentUserId: WalletTestModel.uniqueId);

            // assert
            verify(_mockWalletApi.runSearchQuery(WalletTestModel.uniqueId));
          });

          test('should re-throw a ServerException when server call fails',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenThrow(M4eExceptionMessages.kUndefinedException);

            // act
            final _call = _m4eWalletService.getWallets;

            // assert
            expect(() => _call(currentUserId: WalletTestModel.uniqueId),
                throwsA(TypeMatcher<M4eServerException>()));
          });

          test(
              'should re-throw a ServerException when server call to [WalletApi.runSearchQuery] fails',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => WalletTestModel.uniqueId);
            when(_mockWalletApi.runSearchQuery(any))
                .thenThrow(M4eExceptionMessages.kUndefinedException);

            // act
            final _call = _m4eWalletService.getWallets;

            // assert
            expect(() => _call(currentUserId: WalletTestModel.uniqueId),
                throwsA(TypeMatcher<M4eServerException>()));
          });
        });

        group('other user wallets', () {
          setUp(() {
            when(_mockUserApi.getM4eUser(any))
                .thenAnswer((_) async => WalletTestModel.m4eOtherUser);
          });

          test(
              'should call [WalletApi.queryUserWallets] with appropriate parameters (query with multiple-search-clause when [currentUserId, otherUserId] is passed in)',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => WalletTestModel.uniqueId);

            // act
            final _result = await _m4eWalletService.getWallets(
                currentUserId: WalletTestModel.uniqueId,
                otherUserPhoneNumber: WalletTestModel.phoneNumber);

            // assert
            expect(_result, equals(WalletTestModel.tWalletList));
            verify(_mockWalletApi.queryUserWallets(
                WalletTestModel.multipleSearchFormTypeCollection));
          });
        });
      });

      group('isDisconnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => false);
        });
        test(
            'should throw [ServerException] with [NO_INTERNET_CONNECTION] error code',
            () async {
          // arrange
          when(_mockWalletApi.queryUserWallets(any))
              .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

          // act
          final _call = _m4eWalletService.getWallets;

          // assert
          expect(() => _call(currentUserId: WalletTestModel.uniqueId),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });
    });

    group('getWalletBalance', () {
      test('should verify device connectivity status', () async {
        // arrange
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
        // act
        await _m4eWalletService.getWalletBalance(WalletTestModel.uniqueId);

        // assert
        verify(_mockDataConnectionChecker.hasConnection);
      });

      group('isConnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => true);
        });

        test('should call [M4eWalletApi] to get wallet-balance with wallet-id',
            () async {
          // assert
          when(_mockWalletApi.getWalletBalance(any))
              .thenAnswer((_) async => WalletTestModel.tWalletBalance);

          // act
          final _result = await _m4eWalletService
              .getWalletBalance(WalletTestModel.uniqueId);

          // assert
          verify(_mockWalletApi.getWalletBalance(WalletTestModel.uniqueId));
          expect(_result, equals(WalletTestModel.tWalletBalance));
        });

        test('should re-throw a ServerException when server call fails',
            () async {
          // arrange
          when(_mockWalletApi.getWalletBalance(any))
              .thenThrow(M4eExceptionMessages.kUndefinedException);

          // act
          final _call = _m4eWalletService.getWalletBalance;

          // assert
          expect(() => _call(WalletTestModel.uniqueId),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });

      group('isDisconnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => false);
        });
        test(
            'should throw [ServerException] with [NO_INTERNET_CONNECTION] error code',
            () async {
          // arrange
          when(_mockWalletApi.getWalletBalance(any))
              .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

          // act
          final _call = _m4eWalletService.getWalletBalance;

          // assert
          expect(() => _call(WalletTestModel.uniqueId),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });
    });

    group('createWallet | ', () {
      test('should verify device connectivity status', () async {
        // arrange
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
        // act
        await _m4eWalletService.createWallet(
            WalletTestModel.uniqueId, WalletTestModel.tM4eWalletForm);

        // assert
        verify(_mockDataConnectionChecker.hasConnection);
      });

      group('isConnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => true);
        });

        test('should call [M4eWalletApi] to createWallet with right parameters',
            () async {
          // assert
          when(_mockWalletApi.createWallet(any, any))
              .thenAnswer((_) async => WalletTestModel.uniqueId);

          // act
          final _result = await _m4eWalletService.createWallet(
              WalletTestModel.uniqueId, WalletTestModel.tM4eWalletForm);

          // assert
          verify(_mockWalletApi.createWallet(
              WalletTestModel.authToken.id, WalletTestModel.tM4eWalletForm));
          expect(_result, equals(WalletTestModel.uniqueId));
        });

        test('should re-throw a ServerException when server call fails',
            () async {
          // arrange
          when(_mockWalletApi.createWallet(any, any))
              .thenThrow(M4eExceptionMessages.kUndefinedException);

          // act
          final _call = _m4eWalletService.createWallet;

          // assert
          expect(
              () => _call(
                  WalletTestModel.uniqueId, WalletTestModel.tM4eWalletForm),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });

      group('isDisconnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => false);
        });
        test(
            'should throw [ServerException] with [NO_INTERNET_CONNECTION] error code',
            () async {
          // arrange
          when(_mockWalletApi.createWallet(any, any))
              .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

          // act
          final _call = _m4eWalletService.createWallet;

          // assert
          expect(
              () => _call(
                  WalletTestModel.uniqueId, WalletTestModel.tM4eWalletForm),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });

      test('should call wallet-api with approprite paraameters', () {
        final _permissions = <M4eAccessPermissionSet>[
          EntityWalletAccess.see,
          EntityWalletAccess.view,
          EntityWalletAccess.pay,
          EntityWalletAccess.grant,
          EntityWalletAccess.alter,
          EntityWalletAccess.create,
          EntityWalletAccess.destroy,
        ];

        String _permissionUrl =
            '/wallets/access?entity=${WalletTestModel.uniqueId.toString()}';

        for (final item in _permissions) {
          _permissionUrl += '&access=${item.toString()}';
        }

        print(_permissionUrl);
      });
    });
  });
}
